# -*- coding: cp1252 -*-
"""
    special_elevationmodel.py 
    
    Goal: create elevation model based on specified input, including 
    non-standard Baseline feature classes kades and kribben
        
    
"""
import os
import sys
import math
import gc
import traceback
import datetime as dt
import arcpy
from collections import OrderedDict
import numpy as np

arcpy.env.overwriteOutput = 1
arcpy.env.XYTolerance = 0.001

TOLERANCEDISTANCE = 5.0
SEARCHDISTANCE = 0.01
EQUALITYTOLERANCE = 0.01
EPSILON = 1e-05
ROUNDDECIMALS = 8
LOG_FILE = None


def featCount(ds):
    """ Count table in input table (table, feature class, feature layer)
    
    Arguments:
        ds: ArcGIS table
    
    Returns: 
        integer: number of rows
    """
    result = arcpy.GetCount_management(ds)
    count = int(result.getOutput(0))
    return count


def AddMessage(mes, in_log=True):
    """ Prints message to dialog (being console, ArcTool dialog 
    
    Arguments:
        mes: string containing message
        in_log:     boolean: also add text in log file
    
    Returns: 
        -
        
    Note that print(mes) is disabled, as enabling this results in an
    error in ArcGIS 10.1 when many messages are printed in the ArcTool dialog
    """
    now = dt.datetime.now()
    strNow = now.strftime('%H:%M:%S')
    strMes = '{time} - INFO: {mes}'.format(time=strNow, mes=mes)
    arcpy.AddMessage(strMes)
    print(strMes)    

    if in_log:
        if LOG_FILE:
            LOG_FILE.write(strMes + '\n')
    pass

    
def AddError(mes, in_log=True):
    """ Prints message to dialog (being console, ArcTool dialog 
    
    Arguments:
        mes: string containing message
        in_log:     boolean: also add text in log file
    
    Returns: 
        -
        
    Note that print(mes) is disabled, as enabling this results in an
    error in ArcGIS 10.1 when many messages are printed in the ArcTool dialog
    """
    now = dt.datetime.now()
    strNow = now.strftime('%H:%M:%S')
    strMes = '{time} - ERROR: {mes}'.format(time=strNow, mes=mes)    
    arcpy.AddError(strMes)
    print(strMes)
    
    if in_log:
        if LOG_FILE:
            LOG_FILE.write(strMes + '\n')
    pass
   
   
def AddWarning(mes, in_log=True):
    """ Prints warning to dialog (being console, ArcTool dialog 
    
    Arguments:
        mes:        string containing message
        in_log:     boolean: also add text in log file
    
    Returns: 
        -
        
    Note that print(mes) is disabled, as enabling this results in an
    error in ArcGIS 10.1 when many messages are printed in the ArcTool dialog
    """
    now = dt.datetime.now()
    strNow = now.strftime('%H:%M:%S')
    strMes = '{time} - WARNING: {mes}'.format(time=strNow, mes=mes)      
    arcpy.AddWarning(strMes)
    print(strMes)    
    if in_log:
        if LOG_FILE:
            LOG_FILE.write(strMes + '\n') 
    pass

    
def is_between(firstPoint, lastPoint, betweenPoint):
    """ Analysis wether given point is between provided first- and last point 
    
    Arguments:
        firstPoint:     (x, y) tuple
        lastPoint       (x, y) tuple
        betweenPoint:   (x, y) tuple
    
    Returns: 
        Boolean: true = in between, false = not in between
       
    """
    
    # Bepaal cross-product
    bpx, bpy = betweenPoint[0], betweenPoint[1]
    fpx, fpy = firstPoint[0], firstPoint[1]
    lpx, lpy = lastPoint[0], lastPoint[1]
    dxc = bpx - fpx
    dyc = bpy - fpy
    
    dxl = lpx - fpx
    dyl = lpy - fpy
    cross = dxc * dyl - dyc * dxl;   

    if abs(cross) < EQUALITYTOLERANCE:
        # Dan ligt het punt op de lijn
        if (abs(dxl) >= abs(dyl)):
            if (fpx <= bpx <= lpx) or (lpx <= bpx <= fpx):
                return True
        else:
            if (fpy <= bpy <= lpy) or (lpy <= bpy <= fpy):
                return True   
    return False
    
def length(x0,y0, x1,y1):
    """ Calculate length between first and last points
    
    Arguments:
        x0:             x coordinate of first point
        y0:             x coordinate of last point
        x1:             y coordinate of first point
        y1:             y coordinate of last point
    
    Returns: 
        float:      distance
       
    """
    distX =x1 - x0
    distY = y1 - y0
    return math.sqrt((distX**2 + distY**2))  
         

def unique_values(table, field):
    """ Make an unordered set of unique values in a field
    
    Arguments:
        table:      string: ArGIS table
        field:      string: name of table to be analysed
    
    Returns: 
        unique_set: Python set containing unique values.
        
    Note: sets are unordered. Advantage over list is the low memory print
    """        
    with arcpy.da.SearchCursor(table, [field]) as cursor:
        return sorted({row[0] for row in cursor})    

        
def lines_per_nummer(table, wc, extraFields):
    """ Make an unordered set of unique values in a field
    
    Arguments:
        table:      string: ArGIS table
        field:      string: name of table to be analysed
    
    Returns: 
        unique_set: Python dictionary containing unique values.
        
    """    
    d = {}
    extra = {}
    
    curFields = ['Shape@', 'NUMBER']
    if extraFields:
        curFields = curFields + extraFields
    cursor = arcpy.da.SearchCursor(table, curFields, where_clause=wc)
    for row in cursor:
        key = row[1]
        shape = row[0]
        e = d.get(key, [])
        e.append(shape)
        d[key] = e
        
        if extraFields:
            ext = []
            if len(ext) == 0:
                for i in range(2, len(curFields)):
                    ext.append(row[i])
            extra[key] = ext
                
    del cursor, wc
    return [d, extra]


def getNonZeros(fpArray, lpArray, pointArray):

    # Calculate segment distance (x and y direction)
    segx = (lpArray[:,0] - fpArray[:,0])
    segy = (lpArray[:,1] - fpArray[:,1])
    
    # Define number of points to be analysed
    pointCount = pointArray.shape[0]  
    # Find touching segments
    def findTouchingSegments(segx, segy, pointArray, startSlice=None, endSlice=None):
        # Start with cross-product
        cp = abs(((pointArray[startSlice:endSlice, None, 1] - fpArray[:,1]) * segx) - 
                ((pointArray[startSlice:endSlice, None, 0] - fpArray[:,0]) * segy))
        tf1 = cp > EPSILON

        del cp
        # tf = array waarin wordt bepaald op welk segmenten een of meerdere punten ligt
        tf = np.ones(np.shape(tf1))
        
        tf[tf1==True]=0
        del tf1

        dotproduct = ((pointArray[startSlice:endSlice, None, 0] - fpArray[:,0]) * segx + 
                     (pointArray[startSlice:endSlice, None, 1] - fpArray[:,1]) * segy) 
        tf2 = dotproduct < 0
        tf[tf2==True]=0
        del tf2

        lijnlengte = (segx**2 + segy**2)
        tf3 = dotproduct > lijnlengte
        del dotproduct, segx, segy
        tf[tf3==True]=0
        del tf3

        # In rijen: alle punten
        # In kolommen: het segment waar punt op ligt
        
        unique_items = list(np.argwhere(tf==1)[:,0] + startSlice) 
        del tf
        return unique_items
        
    try:
        allNonZeros = findTouchingSegments(segx, segy, pointArray, startSlice=0, endSlice=pointCount+1)
        allNonZeroSet = set(allNonZeros) 
        del allNonZeros
        return  allNonZeroSet 
    except MemoryError:
        gc.collect()
        allNonZeros = []
        i = 0
        while i <= pointCount:
            j = i + 50
            ss = findTouchingSegments(segx, segy, pointArray, startSlice=i, endSlice=j)
            allNonZeros += ss
            i = j  
        allNonZeroSet = set(allNonZeros) 
        del allNonZeros
        return  allNonZeroSet 


def DefineNewLines(line):
    lDict = OrderedDict()
    includedDict = {}
    prtCount = 0
    
    for shape in line:
        
        for prt in shape.getPart():
            xyList = []
            for pnt in prt:
                xy = [pnt.X, pnt.Y, -9999]
                xyList.append(xy)

            lDict[prtCount] = xyList
            includedDict[prtCount] = False
            
            prtCount += 1
            
    def addLines(nlNum, lineDict, inclDict, nLines):
        newline = nLines[nlNum]
        
        lineChanged = False
        
        for i in lineDict:
            item = lineDict[i]
            if not inclDict[i]:
                x1 = newline[-1][0]
                y1 = newline[-1][1]
                x2 = item[0][0]
                y2 = item[0][1]
                dist = math.sqrt((x1-x2)**2 + (y1-y2)**2)
                if dist < 0.001:
                    inclDict[i] = True
                    # Voeg lijn toe aan newline
                    nLines[nlNum] = newline + item
                    newline = nLines[nlNum]
                    lineChanged = True
        return [lineChanged, nLines]                  
            
    newLines = OrderedDict()                
    nlNum = 0
    joining = True
    while joining:
        joining = False
        for itemNum1 in lDict:
            item1 = lDict[itemNum1]
            fp1 = item1[0]
            lp1 = item1[-1]
            basis = itemNum1
            included1 = includedDict[itemNum1]
            if included1:
                continue
            else:
                for itemNum2 in lDict:
                    if itemNum2 == itemNum1:
                        continue
                    item2 = lDict[itemNum2];
                    included2 = includedDict[itemNum2]
                    if included2:
                        continue
                    else:
                        fp2 = item2[0]
                        lp2=  item2[-1]           
                        dist = math.sqrt((fp1[0] - lp2[0])**2 + (fp1[1] - lp2[1])**2)

                        if dist < 0.001:
                            basis = itemNum2
                            break
                includedDict[basis] = True
                newLine = lDict[basis]
                nlNum += 1
                newLines[nlNum] = newLine

                stillAdding = True
                while stillAdding:
                    added = addLines(nlNum, lDict, includedDict, newLines)
                    newLines = added[1]
                    stillAdding = added[0]
                    joining = True
            if joining:
                break  
    del lDict, includedDict
    return newLines
  
  
def analyzeLine(lijnen, punten, outlines, wc, lijnc, fieldname, sr, extraFields=None):
    # Script loopt random door alle unieke lijnen, omdat dit geen ordereddict is
    #  maar een 'gewone dictionary', waarin de key-value pairs een random volgorde hebben
    
    lineCreated = False
 
    lineData =  lines_per_nummer(lijnen, wc, extraFields)
    unique_lijnen = lineData[0]
    extraFieldData = lineData[1]
    
    icFields = ['Shape@', 'NUMBER']
    if extraFields:
        icFields = icFields + extraFields
    ic = arcpy.da.InsertCursor(outlines, icFields)

    for lijnNummer, lijn in unique_lijnen.iteritems():
        try:
            if (lijnc % 1000) == 0 and lijnc > 0:
                AddMessage('{c} lines processed'.format(c=lijnc),in_log=False)
            
#            if lijnc == 10000:
#                break
            lijnc += 1

            # Maak van lijnen met hetzelfde nummer zo lang mogelijke single-part delen. 
            newLines = DefineNewLines(lijn)
            del lijn
            
            pc = arcpy.da.SearchCursor(punten, ['Shape@XY', fieldname, 'NUMBER'], where_clause='NUMBER = {0}'.format(lijnNummer))
            #plist = [[row[0], row[2]] for row in pc]
            pDict = OrderedDict()        
            pL = []
            pLAppend = pL.append
            getPdict = pDict.get
            
            for row in pc:
                pnt = row[0]
                z = row[1]
                pVal = getPdict(pnt)
                if pVal:
                    if pVal[1] > z:
                        z = pVal[1]
                else:
                    pLAppend([pnt[0], pnt[1]])
                pDict[pnt] = [pnt, z]   
            
            if len(pL) == 0:
                AddWarning('No points found on line with NUMBER = {n}'.format(n=lijnNummer))
                del pDict, newLines, pL, pc
                continue
            
            pl = np.array(pL)
            del pL
            
            updatedZValues = {} 
            allAr = arcpy.Array() 
            arNum = 0
            for key, newline in newLines.iteritems():

                fpl = []    
                lpl = []
                pold = None
                for i in range(0, len(newline)):
                    p = newline[i]
                    if p == pold:
                        continue
                    
                    if i == 0:
                        fpl.append([p[0], p[1]])
                    elif i == len(newline) - 1:
                        lpl.append([p[0], p[1]])
                    else:
                        lpl.append([p[0], p[1]])
                        fpl.append([p[0], p[1]])                
                    pold = p
                fpn = np.array(fpl)
                lpn = np.array(lpl)
                del fpl, lpl

                nonzeros = getNonZeros(fpn, lpn, pl)  
                del fpn, lpn                
                    
                if not nonzeros:
                    continue
                pointsOnLine = {}
                
                pIndex = 0
                for p, pp in pDict.iteritems():
                    if pIndex in nonzeros:
                        pointsOnLine[p] = pp
                    pIndex += 1
                    
                pointCount = len(pointsOnLine)  

                if pointCount == 0:
                    AddWarning('No points found on (one of the) line(s) WITH NUMBER = {0}'.format(lijnNummer))
                    continue
                else:       
                    pointTaken = []
                    
                    listar = []
                    for i in range(0, len(newline)):
                        pnt = newline[i]
                                                       
                        rpoint = (pnt[0], pnt[1])
                        
                        if rpoint in pointsOnLine:
                            z = pointsOnLine[rpoint][1]
                            if not rpoint in pointTaken:
                                pointTaken.append(rpoint)                    
                        else:
                            z = -9999       
                        listar.append([pnt[0], pnt[1], z])
                        
                    polSet = set(list(pointsOnLine.keys()))
                    pointsNoTakenYet = polSet - set(pointTaken)
                    
                    for point in pointsNoTakenYet:
                        for i in range(1, len(listar)):
                            pStart = listar[i-1]
                            pEnd = listar[i]
                            pBetween = pointsOnLine[point][0]   
                            if is_between(pStart, pEnd, pBetween):                    
                                segment_length = length(pStart[0], pStart[1], pEnd[0], pEnd[1])
                                betw_length = length(pStart[0], pStart[1], pBetween[0], pBetween[1])
                                if betw_length < segment_length:
                                    listar.insert(i, [pBetween[0], pBetween[1], pointsOnLine[point][1]])
                                    break
                                        
                    noZVals = []                               
                    ar = arcpy.Array()

                    arCount = 0
                    for i in range(0, len(listar)):
                        curItem = listar[i]
                        point = arcpy.Point(curItem[0], curItem[1], curItem[2])
                        ar.append(point)
                        arCount += 1
                        if curItem[2] == -9999:
                            noZVals.append(i)
                    
                    # Bepaal eerst hoogte van eerste punt als hier geen hoogte-info was
                    #   pak hiervoor hoogte van eerstvolgende punt dat wel hoogte bezit 
                    if len(noZVals) > 0 and noZVals[0] == 0:
                        curObject = ar.getObject(noZVals[0])
                        for i in range(1, arCount):
                            nitem = ar.getObject(i)
                            if nitem.Z > -9999:
                                curround = (curObject.X, curObject.Y)
                                if curround in updatedZValues:
                                    zitem = updatedZValues[curround]
                                    z = zitem[0]
                                    if z < nitem.Z:
                                        z = nitem.Z
                                    zdict = zitem[1]
                                    if arNum in zdict:
                                        zdict[arNum].append(0)
                                    else:
                                        zdict[arNum] = [0] #Z en index van punt in array
                                    updatedZValues[curround] = [z, zdict]
                                else:
                                    updatedZValues[curround] = [nitem.Z, {arNum:[0]}] #Z en index van punt in array
                                break
                        del noZVals[0]   
                        
                    # Bepaal dan hoogte van laatste punt als hier geen hoogte-info was
                    #   pak hiervoor hoogte van voorgaand punt dat wel hoogte bezit            

                    if len(noZVals) > 0 and (noZVals[-1] == (arCount - 1)):
                        curObject = ar.getObject(noZVals[-1])
                        
                        for i in range(arCount -1, -1, -1):
                            nitem = ar.getObject(i)
                            if nitem.Z > -9999:
                                curround = (curObject.X, curObject.Y)
                                if curround in updatedZValues:
                                    zitem = updatedZValues[curround]
                                    z = zitem[0]
                                    if z < nitem.Z:
                                        z = nitem.Z
                                    zdict = zitem[1]
                                    if arNum in zdict:
                                        zdict[arNum].append(noZVals[-1])
                                    else:
                                        zdict[arNum] = [noZVals[-1]] #Z en index van punt in array
                                    updatedZValues[curround] = [z, zdict]
                                else:
                                    updatedZValues[curround] = [nitem.Z, {arNum:[noZVals[-1]]}] #Z en index van punt in array
                                
                                break
                        del noZVals[-1] 
                        
                    # En daarna tussenliggende punten en interpoleer deze uit
                    for index in noZVals:
                        # Bepaal eerst vorige hoogte
                        curObject = ar.getObject(index)
                        curround = (curObject.X,curObject.Y)
                        prevObjectFound = False
                        distFrom = 0
                        for i in range(index - 1, -1, -1):
                            
                            prevObject = ar.getObject(i)
                            distFrom += length(prevObject.X, prevObject.Y, curObject.X, curObject.Y)
                            if prevObject.Z > -9999:
                                prevObjectFound = True
                                break

                        nextObjectFound = False
                        distTo = 0
                        for i in range(index + 1, arCount):
                            nextObject = ar.getObject(i)
                            distTo += length(nextObject.X, nextObject.Y, curObject.X, curObject.Y)
                            if nextObject.Z > -9999:
                                nextObjectFound = True
                                break
                                
                        if nextObjectFound and prevObjectFound:
                            totalDist = distFrom + distTo
                            perc = distFrom / totalDist
                            z = ((nextObject.Z - prevObject.Z) * perc) + prevObject.Z
                        elif nextObjectFound:
                            z = nextObject.Z
                        elif prevObjectFound:
                            z = prevObject.Z
                        else:
                            z = -9999
                        if curround in updatedZValues:
                            zitem = updatedZValues[curround]
                            zold = zitem[0]
                            if zold > z:
                                z = zold
                            zdict = zitem[1]
                            if arNum in zdict:
                                zdict[arNum].append(index)
                            else:
                                zdict[arNum] = [index] #Z en index van punt in array
                            updatedZValues[curround] = [z, zdict]
                        else:
                            updatedZValues[curround] = [z, {arNum: [index]}] #Z en index van punt in array
                    allAr.append(ar)
                    arNum += 1
            
            del newLines, pl
            if allAr.count > 0:      
                for item in updatedZValues:
                    curItem = updatedZValues[item]
                    curZ = curItem[0]
                    
                    if curZ == -9999:
                        AddWarning('Elevation cannot be determined for one or more vertices on line with NUMBER = {0}. Set to -9999'.format(lijnNummer))
                    else:
                        for curAr in curItem[1]:
                            ar = allAr.getObject(curAr)
                            for index in curItem[1][curAr]:
                                curObject = ar.getObject(index)
                                curObject.Z = curZ
            
                polyline = arcpy.Polyline(allAr, sr, True)  
                
                if extraFields:
                    extData = extraFieldData[lijnNummer]
                    ir = [polyline, lijnNummer] + extData
                    ic.insertRow(ir)
                else:
                    ic.insertRow([polyline, lijnNummer])
                lineCreated = True
            if (lijnc % 25000) == 0:
                collected = gc.collect()
        except:
            AddError('Unexpected error while processing line with NUMBER = {0}'.format(lijnNummer))
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)

            err = "\nErrors occured during execution.\nTraceback info:\n" + "".join(tbinfo) + "\nError Info:\n" + str(sys.exc_info()[:2])  
            AddError(err)   
            lineCreated = False
            break        
    del unique_lijnen
    return lineCreated

    
def create3dlinefc(lines, points, fieldname, outlines, template, extraFields=None):
    AddMessage('Converting {fc} to 3D-lines'.format(fc=os.path.basename(lines)))
    lineCreated = False

    AddMessage('Place lines in memory for optimal performance', in_log=False)
    rmem = 'in_memory/lines'
    arcpy.CopyFeatures_management(lines, rmem)
    
    AddMessage('Collect unique NUMBERs', in_log=False)
    unique_lijnen = unique_values(rmem, 'NUMBER')
    
    gc.collect()
    
    sr = arcpy.Describe(lines).spatialReference
        
    arcpy.CreateFeatureclass_management(os.path.dirname(outlines), os.path.basename(outlines),  'POLYLINE', has_z='ENABLED', template=template)
    
    # Loop elk NUMMER
    lijnc = 1
    AddMessage('Start looping lines'.format(fc=os.path.basename(lines)), in_log=False)
    
    nStep = 25000
    for i in range(0, len(unique_lijnen), nStep):
        subList = unique_lijnen[i: i + nStep]
        nMin = min(subList)
        nMax = max(subList)
        wc = 'NUMBER >= {min} AND NUMBER <= {max}'.format(min=nMin, max=nMax)  
        lineCreated = analyzeLine(rmem, points, outlines, wc, i, fieldname, sr, extraFields)

    arcpy.Delete_management(rmem)    
    if not lineCreated:
        pass
        #arcpy.Delete_management(outlines)    
    return lineCreated     



