# -*- coding: cp1252 -*-
"""
    Script: bim6_CreateEditableLinesAndPoints.py
    Version: zie script bim6_versioninfo.py
    Goal: conversion of 3D-lines, (3D)-routes + events to lines and points 
        which can be used for Baseline measure editing

    Author: Koen Volleberg, Lievense
    Voor: Stichting Deltares

    Author: Jan den Daas, WSP
    Voor: Stichting Deltares
    
    Updates:
    Version 0.0.1
        Initial conversion from Baseline 5 protocol to Baseline 6
            Updates:
    Version 0.0.2
        Added batch functionality
"""

import bim6_versioninfo as versioninfo
scriptversion = versioninfo.version

import os, sys
from collections import OrderedDict
import arcpy
arcpy.env.overwriteOutput = True


class FeatureClassToConvert():
    events = None
    eventFields = None
    isValid = True
    
    dictLineTypes = {
        'terrain_edge_3d_lines': '3dline',
        'terrain_jump_3d_routes': '3droute',
        'elevated_line_routes': 'route',
        'bridge_routes': 'route',
    }
    
    dictFields = {
        'elevated_line_events': ['Shape@XY', 'NUMBER', 'MEASURE', 'ELEVATION_CREST', 'ELEVATION_LEFT', 'ELEVATION_RIGHT', 'SLOPE_LEFT', 
                         'SLOPE_RIGHT','CREST_WIDTH', 'ROUGHNESS_CODE'],
        'terrain_jump_3d_events': ['Shape@XY',  'NUMBER', 'MEASURE', 'ELEVATION_LEFT', 'ELEVATION_RIGHT', 'SLOPE_LEFT', 
                         'SLOPE_RIGHT','CREST_WIDTH', 'ROUGHNESS_CODE'],
        'terrain_edge_3d_lines': ['Shape@', 'NUMBER', 'CHARACTERISTICS'],
        'terrain_jump_3d_routes': ['Shape@', 'NUMBER', 'CHARACTERISTICS'],
        'elevated_line_routes': ['Shape@', 'NUMBER', 'TYPE', 'CHARACTERISTICS'],
        'bridge_routes': ['Shape@', 'NUMBER', 'NAME', 'CHARACTERISTICS'],
        'bridge_events': ['Shape@XY', 'NUMBER', 'MEASURE','NAME', 'CHARACTERISTICS', 'DIAMETER', 'CP']
                         }
                     
    def __init__(self, inFeatures):
        desc = arcpy.Describe(inFeatures)
        self.name = desc.basename.lower()
        
        self.type = self.dictLineTypes.get(self.name, None)
        self.fields = self.dictFields.get(self.name, None)
        self.location = desc.path
        self.fullPath = os.path.join(self.location, self.name)
        self.fgdb = os.path.dirname(self.location)        
        if self.type == '3dline':
            self.hasEvents = False
        elif self.type == None:
            self.hasEvents = False
            self.isValid = False
            return
        else:
            self.hasEvents = True
            eventsName = self.name.replace("routes","events")    
            self.events = os.path.join(self.fgdb, eventsName)                 
            self.eventFields = self.dictFields[eventsName]
                
        if self.type == '3dline':
            subName = self.name[:-9]
        elif self.type == '3droute':
            subName = self.name[:-10]
        else: 
            subName = self.name[:-7]
            
        #route en punten wegschrijven in gdb
        self.outputPointName = subName + "_points"
        self.outputLineName = subName + "_lines"
        self.outputTempName = subName + "_temp"
        self.pointFc = os.path.join(self.location, self.outputPointName)
        self.lineFc = os.path.join(self.location, self.outputLineName)
        self.tempFc = os.path.join(self.location, self.outputTempName)               

def booleanizeInput(inputParameter):
    if inputParameter == True or inputParameter == False:
        return inputParameter
    elif str(inputParameter) == '1' or str(inputParameter.lower()) == 'true':
        return True
    else:
        return False   

            
def AddMessage(strMessage):
    ''' print and AddMessage messages in one routine
    '''
    print(strMessage)
    arcpy.AddMessage(strMessage)


def AddError(err):
    ''' print and AddError error messages in one routine
    '''
    print(err)
    arcpy.AddError(err)


def AddWarning(war):
    ''' print and AddError error messages in one routine
    '''
    print(war)
    arcpy.AddWarning(war)

        
def convertRoute(feat2convert,swan=False):
    '''
    Convert 2D routes, ignoring any Z-values in the geometry attribute,
    and store event data as point feature class
    
    TODO: 
    Two checks are built in, in this order (tolerance distance = 1mm):
    1. events without a vertex: an extra vertex is added at this location so event and vertices are consistent
    2. vertices without an event: an extra poiny will be placed with default values
        
    Assumption: input is single-part as this is conform Baseline protocol
    '''
    if not feat2convert.hasEvents:
        AddError('Selected feature class does not have corresponding events.')
        return
        
    AddMessage("Start conversion of normal(2D) routes and events")   
    
    # Convert routes to lines (sipmle copy features)
    AddMessage("Converting routes...")
    
    '''
    lineSc = arcpy.da.SearchCursor(feat2convert.fullPath, feat2convert.fields)
    
    # Dictionary met key=NUMMER en value is list of lists with:
    #    (Shape@, type, chars, list with rounded XY-values)
    dictLines = {}
    
    for row in lineSc:
        number = row[1]
        shape = row[0]
        type = row[2]
        chars = row[3]
        
        prt = feat.getPart(0)
        
        if number in dictLines:
            AddWarning('Multiple lines with same number [{nr}] are found. Please fix before finalizing Baseline measure.'.format(nr=number))
        else:
            dictLines[number] = []
                        
        coordArray = []                
        for pnt in prt:
            xR = round(pnt.X, 3)
            yR = round(pnt.Y, 3)
            xy = (xR, yR)
            coordArray.append(xy)
        dictLines[number].append([shape, type, chars, coordArray])
    '''                              
    arcpy.CopyFeatures_management(feat2convert.fullPath, feat2convert.lineFc)

    # Conversion of events
    AddMessage("Converting events...")
    pointLayer = "pointLayer"   
    arcpy.MakeRouteEventLayer_lr(feat2convert.fullPath, "NUMBER", feat2convert.events, "NUMBER POINT MEASURE", pointLayer)
    arcpy.CreateFeatureclass_management(feat2convert.location, feat2convert.outputPointName, "POINT", pointLayer) 
    
    # Kopieren naar database van event layer. Noodzakelijk om de events te kunnen sorteren in searchcursor
    # Enige manier om een gesorteerde (obv NUMBER en MEASURE) feature class te krijgen
    arcpy.CopyFeatures_management(pointLayer, feat2convert.tempFc)      

    fields = feat2convert.eventFields    
    ic = arcpy.da.InsertCursor(feat2convert.pointFc, field_names=fields)
    sc = arcpy.da.SearchCursor(feat2convert.tempFc, field_names=fields, sql_clause=(None, "ORDER BY NUMBER, MEASURE"))

    
    for event in sc:
        index = 0
        
        '''
        line =  dictLines.get(event[1],None)
        if not line:
            AddWarning('Event with NUMBER = {nr} does not have a corresponding line. Event is not converted'.format(nr=event[1]))
            continue
        
        shape = event[0]
        rxy = (round(shape.X, 3), round(shape.Y, 3))
        for unLine in line:
            cAr = unLine[3]
            lShape = unLine[0]
            if xy.within(lShape, 'BOUNDARY'):
                if not rxy in cAr:
                    # De event ligt op de lijn, maar er is geen vertex.
                    # De shape van de lijn moet worden uitgebreid met deze X- en Y-coordinaten
                
        '''    
              
        newRow = [None]*len(fields)
        for field in fields:
            newRow[index] = event[index]
            index += 1
        ic.insertRow(newRow)
    arcpy.DeleteField_management(feat2convert.pointFc, "MEASURE")
    del ic, sc
    if not swan:
        AddMessage("Delete original and intermediate data")
        arcpy.Delete_management(feat2convert.fullPath)
        arcpy.Delete_management(feat2convert.events)
    arcpy.Delete_management(feat2convert.tempFc)
    arcpy.Delete_management(pointLayer)
    AddMessage("Finished converting routes and events")

    return [feat2convert.lineFc,feat2convert.pointFc]
    

def convert3DRoute(feat2convert, swan=False):
    '''
    Convert 3D routes, and store event data as point feature class
    '''    
    if not feat2convert.hasEvents:
        AddError('Selected feature class does not have corresponding events.')
        return
        
    AddMessage("Start conversion of 3D routes and events")   
    
    # Convert routes to lines (sipmle copy features)
    AddMessage("Converting routes...")
    
    dictXYZ = OrderedDict()
    arcpy.CreateFeatureclass_management(feat2convert.location, feat2convert.outputLineName, 'POLYLINE', feat2convert.fullPath, has_z='DISABLED')
    fields = feat2convert.fields
    ic = arcpy.da.InsertCursor(feat2convert.lineFc, fields)
    sc = arcpy.da.SearchCursor(feat2convert.fullPath, fields)
    for row in sc:
        index = 0
        nm = row[1]
        newRow = [None]*len(fields)
        for field in fields:
            newRow[index] = row[index]
            index += 1
        ic.insertRow(newRow)
        
        feat = row[0]
        for prt in feat.getPart():
            for pnt in prt:
                if pnt:
                    xyRound = (round(pnt.X, 3), round(pnt.Y, 3))      
                    allInfo = dictXYZ.get(xyRound, [[],[]]) # Als niet bestaat, dan lege lijst met twee lege lijsten maken
                    # Lijst 0 = alle Zwaarden op deze locatie, lijst 1 is alle lijnnummers met vertices op deze locatie
                    if not pnt.Z in allInfo[0]:
                        allInfo[0].append(pnt.Z)  
                    if not nm in allInfo[1]:
                        allInfo[1].append(nm)
                    dictXYZ[xyRound] = allInfo
    del ic    
    
    # Conversion of events
    AddMessage("Converting events...")
    pointLayer = "pointLayer"   
    arcpy.MakeRouteEventLayer_lr(feat2convert.fullPath, "NUMBER", feat2convert.events, "NUMBER POINT MEASURE", pointLayer)
    arcpy.CreateFeatureclass_management(feat2convert.location, feat2convert.outputPointName, "POINT", pointLayer) 
    arcpy.AddField_management(feat2convert.pointFc, 'ELEVATION_CREST',  'DOUBLE')
    # Kopieren naar database van event layer. Noodzakelijk om de events te kunnen sorteren in searchcursor
    # Enige manier om een gesorteerde (obv NUMBER en MEASURE) feature class te krijgen
    arcpy.CopyFeatures_management(pointLayer, feat2convert.tempFc)      

    fields = feat2convert.eventFields   
    newFields = list(fields)
    newFields.append('ELEVATION_CREST')
    ic = arcpy.da.InsertCursor(feat2convert.pointFc, field_names=newFields)

    sc = arcpy.da.SearchCursor(feat2convert.tempFc, field_names=fields, sql_clause=(None, "ORDER BY NUMBER, MEASURE"))
        
    nummerNoVertex = []    
    eventXyDict = {}
    for event in sc:
        index = 0
        num = event[1]
        newRow = [None]*len(newFields)
        for field in fields:
            newRow[index] = event[index]
            index += 1
            
        # Voeg Z toe op de locatie als deze bestaat
        xy = event[0]
        xyRound = (round(xy[0], 3), round(xy[1], 3))    
        
        allInfo = dictXYZ.get(xyRound, None)
        if not allInfo:
            if not num in nummerNoVertex:            
                AddWarning('Line with NUMBER = {nm} has events that are not placed on a vertex, ELEVATION_CREST is set to -999. Please check -999 values'.format(nm=num))
                nummerNoVertex.append(num)
            z = -999
        else:
            allZs = allInfo[0]
            if len(allZs) > 1:
                AddWarning('Multiple, different elevations are found on same location. X:{x}. Y:{y}. Maximum Z-value is converted. Please check data'.format(x=xyRound[0], y=xyRound[1]))
                z = max(allZs)
            elif len(allZs) == 1:
                z = allZs[0]
            else:
                AddWarning('Line with NUMBER = {nm} has events that are not placed on a vertex, ELEVATION_CREST is set to -999. Please check -999 values'.format(nm=num))
                nummerNoVertex.append(num)
                z = -999
        newRow[-1] = z
        ic.insertRow(newRow)
        
        eventXyNums = eventXyDict.get(xyRound, [])
        eventXyNums.append(num)
        eventXyDict[xyRound] = eventXyNums
        
    nummerNoEvent = []    
    for xy in dictXYZ:
        #Check of er uberhaupt geen events zijn op deze vertex
        if not xy in eventXyDict:
            xyNums = dictXYZ[xy][1]
            for num in xyNums:
                if not num in nummerNoEvent:               
                    AddWarning('Line with NUMBER = {num} has vertices that do not match with a corresponding event. Please check for -999 values'.format(num=num))
                    nummerNoEvent.append(num)
                newRow = [xy,  num, 99, -999, -999, -999, -999,-999, 0,-999]
                ic.insertRow(newRow)
        # Als deze er wel zijn, kan het ook nog zo zijn dat ze voor lijnen met bepaald nummer niet bestaan en voor andere wel        
        else:
            # Dus check per nummer of ze er zijn
            lineNums = dictXYZ[xy][1]
            eventNums = eventXyDict[xy]
            for num in lineNums:
                if not num in eventNums:
                    if not num in nummerNoEvent:                   
                        AddWarning('Line with NUMBER = {num} has vertices that do not match with a corresponding event. Please check for -999 values'.format(num=num))
                        nummerNoEvent.append(num)
                    newRow = [xy,  num, 99, -999, -999, -999, -999,-999, 0, -999]
                    ic.insertRow(newRow)              
                                             
    arcpy.DeleteField_management(feat2convert.pointFc, "MEASURE")
    del ic, sc
    if not swan:
        AddMessage("Delete original and intermediate data")
        arcpy.Delete_management(feat2convert.fullPath)
        arcpy.Delete_management(feat2convert.events)
    arcpy.Delete_management(feat2convert.tempFc)
    arcpy.Delete_management(pointLayer)
    AddMessage("Finished converting routes and events")

    return [feat2convert.lineFc,feat2convert.pointFc]
    
    pass

    
def convert3DLine(feat2convert,swan=False):
    '''
    Convert 3D lines to lines and points, storing Z-value as attribute in the point feature class
    '''
    if feat2convert.hasEvents:
        AddError('Selected lines have events, use convertRoute or convert3DRoute instead')
        return
    
    AddMessage("Start conversion of 3D lines to 2D lines")   
    
    arcpy.CreateFeatureclass_management(feat2convert.location, feat2convert.outputLineName, 'POLYLINE', feat2convert.fullPath, has_z='DISABLED')
    fields = feat2convert.fields
    ic = arcpy.da.InsertCursor(feat2convert.lineFc, fields)
    sc = arcpy.da.SearchCursor(feat2convert.fullPath, fields)
    for row in sc:
        index = 0
        newRow = [None]*len(fields)
        for field in fields:
            newRow[index] = row[index]
            index += 1
        ic.insertRow(newRow)
    del ic
    
    AddMessage("Start conversion of 3D lines to 2D points containing ELEVATION_CREST attribute")  
    arcpy.CreateFeatureclass_management(feat2convert.location, feat2convert.outputPointName, "POINT", has_z='DISABLED')
    arcpy.AddField_management(feat2convert.pointFc, 'NUMBER', 'LONG')    
    arcpy.AddField_management(feat2convert.pointFc, 'ELEVATION_CREST', 'DOUBLE')
    
    pointFields = ['Shape@XY', 'NUMBER', 'ELEVATION_CREST']
    
    sc.reset()
    ic2 = arcpy.da.InsertCursor(feat2convert.pointFc, pointFields)
    for row in sc:
        feat = row[0]
        nm = row[1]
        
        for prt in feat.getPart():
            for pnt in prt:
                if pnt:
                    newRow = [(pnt.X, pnt.Y), nm, pnt.Z]
                    ic2.insertRow(newRow)                    
    del sc, ic2                    
    if not swan:
        AddMessage("Delete original and intermediate data")
        arcpy.Delete_management(feat2convert.fullPath)
    AddMessage("Finished converting 3d lines")

    return [feat2convert.lineFc, feat2convert.pointFc]                    
                                                 
          
def CreateEditableLinesAndPoints(inFc, swan=False):
    AddMessage("Script: %s" % os.path.basename(sys.argv[0]))
    AddMessage("Version: %s" % scriptversion)
    AddMessage("WSP / Deltares, 2021\n")
    
    # bepalen locatie workspace en naam type lijnen
    fc2Convert = FeatureClassToConvert(inFc)
    if arcpy.Exists(os.path.join(fc2Convert.location, 'elevation_model_terrain')) and not swan:
        AddError('Selected feature class is part of a variant. This tool is meant for Baseline measures only.')
        return None
    if arcpy.Exists(fc2Convert.pointFc):
        arcpy.Delete_management(fc2Convert.pointFc)
    if arcpy.Exists(fc2Convert.tempFc):
        arcpy.Delete_management(fc2Convert.tempFc)
    if arcpy.Exists(fc2Convert.lineFc):
        arcpy.Delete_management(fc2Convert.lineFc)         
    
    lineType = fc2Convert.type
    
    if lineType == 'route':
        convertRoute(fc2Convert,swan)        
    elif lineType == '3droute':
        convert3DRoute(fc2Convert,swan)
    elif lineType == '3dline':
        convert3DLine(fc2Convert,swan)
    else:
        AddError('Selected feature class is not a valid Baseline 6 feature class')
        return None
        


if __name__ == "__main__":
    inFc = arcpy.GetParameterAsText(0)
    try:
        swan = booleanizeInput(arcpy.GetParameter(1))
    except:
        swan = False
    try:
        batch = booleanizeInput(arcpy.GetParameter(2))
    except:
        batch = False

    if batch:
        fgdbs = os.path.join(inFc, 'baseline.gdb')
        for batchroutes in [r'elevation/terrain_edge_3d_lines',r'elevation/terrain_jump_3d_routes',r'elevation/elevated_line_routes',r'locations/bridge_routes']: 
            if arcpy.Exists(os.path.join(fgdbs,batchroutes)):
                CreateEditableLinesAndPoints(os.path.join(fgdbs,batchroutes), swan=False)
    else:
        CreateEditableLinesAndPoints(inFc, swan)

