import os
import arcpy
import settings
from generalFunctions import *
from os.path import join

"""
    
Baseline 6 Input Module

AIM
    - Places a breakline inside a polygon (most likely a open water body, aka lake)

Lievense 2017
Niels Mulder

Based on bim5_BreuklijnInPlassen.py by Koen Volleberg, Lievense

"""

arcpy.env.overwriteOutput = True

def breaklineInsidePolygon(polygon, bufferDistance, elevation, interval, elevationRelativeToGroundLevel = True, onlyAllowOpenWaterBodies = True, forceInsideBuffer = True):

    # Check for selected. No way someone would like to performe this action in ALL features in a FC
    desc = arcpy.Describe(polygon)

    if not hasattr(desc, 'FIDSet') or desc.FIDSet == None or len(desc.FIDSet) == 0:
        AddError('Make a selection of land_use_polygons in ArcMap before using this tool.')
        return False

    # Check wether only open water bodies are slected
    if onlyAllowOpenWaterBodies:
        with arcpy.da.SearchCursor(polygon, ['OID@', 'ROUGHNESS_CODE']) as cursor:
            for row in cursor:
                if row[1] not in [102, 103, 104, 105, 106, 201, 302, 303, 304, 305]:
                    AddError('OID {} is not an open water body: roughness_code = {}. Make sure you only select open water bodies (roughness_code 102, 103, 104, 105, 106, 201, 302, 303, 304, 305) *OR* uncheck the only-allow-open-water-bodies-checkbox to skip this check.'.format(row[0], row[1]))
                    return

    # Buffer
    if forceInsideBuffer:
        bufferDistance = -1*abs(bufferDistance)
        
    AddMessage('Buffer with {} meter...'.format(bufferDistance))
    bufferDistance = "{} Meters".format(bufferDistance)
    arcpy.Buffer_analysis(polygon, 'in_memory/buffer', bufferDistance)
    
    # Polygon to line
    AddMessage('Polygons to lines...')
    arcpy.FeatureToLine_management('in_memory/buffer', 'in_memory/terrain_edge_lines')
    
    # Find current max number
    linesToEdit = os.path.join(desc.dataElement.catalogPath, '../../elevation/terrain_edge_lines')
    pointsToEdit = os.path.join(desc.dataElement.catalogPath, '../../elevation/terrain_edge_points')
    
    increaseWith = 0
    if arcpy.Exists(linesToEdit):
        with arcpy.da.SearchCursor(linesToEdit, ['NUMBER']) as cursor:
            for row in cursor:
                increaseWith = max(increaseWith, row[0])
    
    # Set number
    AddMessage('Numer lines...')
    AddMessage('Line numers are set to {} + FID.'.format(increaseWith+1))
    arcpy.AddField_management('in_memory/terrain_edge_lines', 'NUMBER', 'SHORT')
    arcpy.CalculateField_management('in_memory/terrain_edge_lines', 'NUMBER', '[OBJECTID] + {}'.format(increaseWith+1), 'VB', '')
    
    # Simplify
    AddMessage('Simplify lines...')
    #SimplifyLine_cartography wont work so we. Error 000207. So we use Generalize_edit
    #arcpy.SimplifyLine_cartography('in_memory/terrain_edge_lines1', 'in_memory/terrain_edge_lines', "POINT_REMOVE", 0.5)
    arcpy.Generalize_edit('in_memory/terrain_edge_lines', interval/10)
    
    # Set points
    AddMessage('Place points on lines...')
    import placePointsOnLine
    placePointsOnLine.placePointOnLine('in_memory/terrain_edge_lines', overwritePoints=True, hoogteofdiameter=elevation, interval=interval)
    
    # Set elevation
    if elevationRelativeToGroundLevel:
        AddMessage('Calculate elevations...')
        with arcpy.da.SearchCursor('in_memory/terrain_edge_lines', ['NUMBER', 'GROUNDLEVEL']) as cursor:
            for row in cursor:
                with arcpy.da.UpdateCursor('in_memory/terrain_edge_points', ['elevation_crest'], where_clause="number = {}".format(row[0])) as cursorUpdate:
                    for row1 in cursorUpdate:
                        if row[1] is None:
                            newElevation = elevation*-1
                        else:
                            newElevation = row[1]-elevation
                        row1[0] = newElevation
                        cursorUpdate.updateRow(row1) 
 
    # Create piont, lines if not existsAppend
    scriptGDB = GetBaselineScriptTemplate()
    if not arcpy.Exists(linesToEdit):
        arcpy.CopyFeatures_management(os.path.join(scriptGDB, 'terrain_edge_lines'), linesToEdit)
    if not arcpy.Exists(pointsToEdit):
        arcpy.CopyFeatures_management(os.path.join(scriptGDB, 'terrain_edge_points'), pointsToEdit)

    # Append
    AddMessage('Save points and lines...')
    arcpy.Append_management('in_memory/terrain_edge_lines', linesToEdit, 'NO_TEST')
    arcpy.Append_management('in_memory/terrain_edge_points', pointsToEdit, 'NO_TEST')
    
    # Delete
    arcpy.Delete_management('in_memory/buffer')
    arcpy.Delete_management('in_memory/terrain_edge_lines')
    arcpy.Delete_management('in_memory/terrain_edge_points')
    

    """
    This procedure is for 3D lines, without points
    # Polygon to line
    AddMessage('Polygons to lines...')
    arcpy.env.outputZFlag = "Enabled"
    arcpy.FeatureToLine_management('in_memory/buffer', 'in_memory/line')
    
    # Simplify
    AddMessage('Simplify lines...')
    arcpy.SimplifyLine_cartography('in_memory/line', '{}_tmp'.format(polygon), "POINT_REMOVE", 0.5)
    
    # Calculate elvation elevation
    AddMessage('Calculate elevations...')
    arcpy.AddField_management('{}_tmp'.format(polygon), 'newElevation', 'DOUBLE')
    arcpy.CalculateField_management('{}_tmp'.format(polygon), 'newElevation', 'a(!groundlevel!, {}, {})'.format(elevation, elevationRelativeToGroundLevel), 'PYTHON', 'def a(b,c,d):\n if not d:\n  return c\n elif b is None:\n  return -c\n return b-c')
    
    # Set elevation
    # We assume the current z value is 0 and the newElevation-field contains the new absolute elevation
    AddMessage('Set elevations...')
    arcpy.Adjust3DZ_management('{}_tmp'.format(polygon), 'NO_REVERSE', 'newElevation')
    
    # Append
    breaklinesFC = os.path.join(desc.dataElement.catalogPath, '../../elevation/terrain_edge_3d_lines')
    if not arcpy.Exists(breaklinesFC):
        arcpy.CopyFeatures_management(os.path.join(GetBaselineMeasureTemplateLocation(), 'baseline.gdb/elevation/terrain_edge_3d_lines'), breaklinesFC)
    arcpy.Append_management('{}_tmp'.format(polygon), breaklinesFC, 'NO_TEST')

    # Delete
    arcpy.Delete_management('in_memory/buffer')
    arcpy.Delete_management('in_memory/line')
    arcpy.Delete_management('{}_tmp'.format(polygon))
    """

if __name__ == "__main__":    
    polygon = arcpy.GetParameterAsText(0)
    bufferDistance = arcpy.GetParameter(1)
    elevation = arcpy.GetParameter(2)
    interval = arcpy.GetParameter(3)
    elevationRelativeToGroundLevel = arcpy.GetParameter(4)  # Defualt = true
    onlyAllowOpenWaterBodies = arcpy.GetParameter(5)  # Defualt = true
    forceInsideBuffer = arcpy.GetParameter(6)  # Defualt = true
    
    #polygon = 'C:\\baseline\\spelen\\testen\\v25\\myBL6collection\\ref\\baseline.gdb\\roughness\\land_use_polygons'
    #bufferDistance = 5
    #elevation = -2
    #interval = 1000
    #elevationRelativeToGroundLevel = True
    #onlyAllowOpenWaterBodies = True
    #forceInsideBuffer = True

    breaklineInsidePolygon(polygon, bufferDistance, elevation, interval, elevationRelativeToGroundLevel, onlyAllowOpenWaterBodies, forceInsideBuffer)
