import os
import arcpy
import settings
from generalFunctions import *
from arcpy.sa import *

"""
    
Baseline 6 Input Module

AIM
    - Remove empty FC

Lievense 2017
Niels Mulder

Based on bim5_createOmtrekMaatregel.py by Koen Volleberg, Lievense

20220216 Jan den Daas
Buffer distance is configurable now and also buffers polygons
"""

arcpy.env.overwriteOutput = 1


def CreateOmtrekMaatregel(inputDir,kenmerk,maatregel):

    AddMessage('Baseline Input Module Tool: {}. Version {}'.format(os.path.basename(sys.argv[0]), settings.getVersion()))
    if not settings.isDef():
        AddMessage('This is a beta version. Do not use in production.')
    AddMessage('WSP {}'.format(settings.getVersionDate()))
    
    fgdb = os.path.join(inputDir, "baseline.gdb")
    wcOmtrekFc = os.path.join(fgdb, "metadata/measure_contour_polygons")
    if arcpy.Exists(wcOmtrekFc):
        AddWarning("Existing measure_contour_polygons will be overwritten...")
        arcpy.Delete_management(wcOmtrekFc)
        
    fileDict = {"poly":[],
                "arc": [],
                "point":[]}

    #Check if sea database
    sea = False
    if arcpy.Describe(os.path.join(inputDir, r"baseline.gdb\elevation")).spatialReference.name == "GCS_WGS_1984":
        sea=True



    # Set workspace
    arcpy.env.workspace = os.path.join(inputDir, 'baseline.gdb')

    # List datasets
    datasets = arcpy.ListDatasets()

    # Add None to datasets. Those points to the files not in a dataset
    datasets = [''] + datasets if datasets is not None else []

    # Loop over datasets
    for dataset in datasets:
        for featureClass in arcpy.ListFeatureClasses(feature_dataset=dataset):
            path = os.path.join(arcpy.env.workspace, dataset, featureClass)
            desc = arcpy.Describe(path)
            typeOfShape = desc.shapeType.lower()
            if typeOfShape == 'polygon':
                fileDict['poly'].append(os.path.join(dataset, featureClass))
            elif typeOfShape == 'polyline':
                fileDict['arc'].append(os.path.join(dataset, featureClass))
            elif typeOfShape == 'point':
                fileDict['point'].append(os.path.join(dataset, featureClass))
            

    

    arcpy.CreateFeatureclass_management(fgdb, "tempWComtrek", "POLYGON")
    tempFc = os.path.join(fgdb, "tempWComtrek")
    arcpy.MakeFeatureLayer_management (tempFc, "tempFcLayer")
    
    
    
    
    #merge extent elevation_raster
    if sea==True and arcpy.Exists("elevation_raster"):
        rasterextent= IsNull("elevation_raster")
        rasterextent.save("rasterextent")
        arcpy.RasterToPolygon_conversion("rasterextent", "rasterpoly", "NO_SIMPLIFY","VALUE")
        arcpy.MakeFeatureLayer_management("rasterpoly", "rasterpolylyr") 
        arcpy.SelectLayerByAttribute_management("rasterpolylyr", "NEW_SELECTION", ' "gridcode" < 1 ')
     

        if arcpy.Exists(tempFc):
            arcpy.Append_management("rasterpolylyr",tempFc,"NO_TEST")
        else:
            arcpy.Merge_management("rasterpolylyr",tempFc)
        arcpy.Delete_management("rasterextent")
        arcpy.Delete_management("rasterpoly")
        arcpy.Delete_management("rasterpolylyr")
        
    else:
        #Merge polygonen
        polyItems = fileDict["poly"]
        polyList = []
        polyString = ""
        for polyItem in polyItems:
            polyPath = os.path.join(fgdb,polyItem)
            if arcpy.Exists(polyPath):
                polyList.append(polyPath)

        if len(polyList)>0:        
            AddMessage("Merge polygons...")
            str1 = ';'.join(polyList)
            arcpy.Append_management(str1,tempFc, "NO_TEST")
        

        tempFcBuf=arcpy.Buffer_analysis(tempFc, tempFc+"_buf","{} Meters".format(distance))
        AddMessage("Buffer polygons ({}m)...".format(distance))
        if arcpy.Exists(tempFc):
            arcpy.Delete_management(tempFc)
        tempFc=tempFcBuf

        #Merge lijnen
        arcItems = fileDict["arc"]
        arcList = []
        for arcItem in arcItems:
            padArc = os.path.join(fgdb, arcItem)
            if arcpy.Exists(padArc):
                fcBuf = padArc + "_buf" 
                arcpy.MakeFeatureLayer_management(padArc,"lyr")
                arcpy.SelectLayerByLocation_management("lyr", "COMPLETELY_WITHIN",tempFc, 0, "new_selection")
                arcpy.SelectLayerByAttribute_management("lyr", "SWITCH_SELECTION")
                arcpy.Buffer_analysis("lyr", fcBuf,"{} Meters".format(distance))
                arcpy.Delete_management("lyr")
                arcList.append(fcBuf)
                
        AddMessage("Combine buffered ({}m) lines...".format(distance))
        arcString = ';'.join(arcList)
        if len(arcList) > 0:
            if arcpy.Exists(tempFc):
                arcpy.Append_management(arcString,tempFc,"NO_TEST")
            else:
                arcpy.Merge_management(arcString,tempFc)
            for arcItem in arcList:
                arcpy.Delete_management(arcItem)

        #Merge punten
        pointItems = fileDict["point"]
        pointList = []

        for pointItem in pointItems:
            padPoint = os.path.join(fgdb, pointItem)
            if arcpy.Exists(padPoint):
                fcBuf = padPoint +"_buf"
                arcpy.MakeFeatureLayer_management(padPoint,"lyr")
                arcpy.SelectLayerByLocation_management("lyr", "INTERSECT",tempFc, 0, "new_selection")
                arcpy.SelectLayerByAttribute_management("lyr", "SWITCH_SELECTION")
                arcpy.Buffer_analysis("lyr", fcBuf,"{} Meters".format(distance))
                arcpy.Delete_management("lyr")
                pointList.append(fcBuf)
        AddMessage("Combine buffered ({}m) points ...".format(distance))


        pointString = ';'.join(pointList)
        if len(pointList) > 0:
            if arcpy.Exists(tempFc):
                arcpy.Append_management(pointString,tempFc,"NO_TEST")
            else:
                arcpy.Merge_management(pointString,tempFc)
            for pointItem in pointList:
                if arcpy.Exists(pointItem):
                    arcpy.Delete_management(pointItem)

    AddMessage("Cleanup intermediate data...")
    if arcpy.Exists(tempFc):
        dis = "in_memory/dis"
        arcpy.Dissolve_management(tempFc,dis,"#","#","SINGLE_PART")
        #arcpy.RepairGeometry_management(dis)
        if not sea:
            if distance < 5:
                simple = float(distance)/5
            else:
                simple = 1
            arcpy.Generalize_edit(dis, "{} Meters".format(simple))
            AddMessage("Generalize polygon with ({}m) increments".format(simple))
        arcpy.Delete_management(tempFc)


        template = GetBaselineMeasureTemplateLocation(sea)
        templateFile = os.path.join(template, 'baseline.gdb/metadata/measure_contour_polygons')


        arcpy.Copy_management(templateFile,wcOmtrekFc)
        arcpy.Append_management(dis,wcOmtrekFc,"NO_TEST")
        arcpy.Delete_management(dis)
        
        arcpy.CalculateField_management(wcOmtrekFc,"CHARACTERISTICS","\"" + kenmerk + "\"")
        arcpy.CalculateField_management(wcOmtrekFc,"MEASURE","\"" + maatregel + "\"")


if __name__ =='__main__':
    inDir = arcpy.GetParameterAsText(0)
    km = arcpy.GetParameterAsText(1)
    maat = arcpy.GetParameterAsText(2)
    distance = arcpy.GetParameter(3)
    CreateOmtrekMaatregel(inDir,km, maat)