# -*- coding: cp1252 -*-
"""

   Script: bim5_placePointOnLine.py

   Versie: Zie script bim5_versipninfo.py

   Geschikt voor ArcGIS 9.3.1 en hoger

   Doel:   plaatsen van punten op hoogtelijnen 

   Auteur: Koen Volleberg, Lievense
   Voor: Stichting Deltares

  Datum: 07022014

  Updates: 
  versie 1  12022014 Koen Volleberg
      - afronding op mm's voor bepaling afstand op lijn

  versie 1.1  12112014 Koen Volleberg
    - tool plaatst altijd punten of vertices op locaties waar een van beide ontbreekt,
         output is altijd een bestand met evenveel lijnen als punten
         AANNAME: bestaande punten moeten op lijnen liggen

  versie 2.3  19102016 Koen Volleberg
      - bugfix: teenhoogte werd niet standaard op -9999 gezet als nieuwe punten
          worden gemaakt voor bandijken

  versie 2.7  21062017 Koen Volleberg
      - bugfix: als een lijn self-intersecting is en twee punten bevat die bewaard moeten worden (bijv lijn om plas)
        werd alleen de eerste bewaard als punten bewaard moesten blijven. Laatste werd weggegooid en vervangen door een nieuwe
      - bugfix: als punten bewaard moeten worden, werd alleen de kruinhoogte bewaard. Nu wordt alle data bewaard.
   
    Version: 2.8
    Datum 20171127
    Door: Niels Mulder
    Update to Baseline 6

    Version: 2.9
    Datum 20180717
    Door: Jan den Daas
    Overgang naar Baseline 6 voltooid.
    Veldnamen verandert regel 398 en 401
    
    Version: 3.0
    Datum 20180723
    Door: Jan den Daas
    Punten worden nu ook overschreven bij een selectie
    Else statement in regel 284-293
    
    Version: 3.1
    Datum 20180723
    Door: Jan den Daas
    Error toegevoegd wanneer er punten buiten lijnen vallen zonder nummer. 
    Dit zorgde voor het vastlopen van de tool 2a en het verwijderen van alle punten.
    
    Version: 3.2
    Datum 20180723
    Door: Jan den Daas
    Bugs gefixt die ontstaan zijn door de introductie van bridges

    Version: 3.3
    Datum 20190315
    Door: Jan den Daas
    Bugs gefixt uit Testreport Baseline 6.1.1.2033
    
    Version: 3.4
    Datum 20200423
    Door: Jan den Daas
    Bugs gefixt m.b.t existing terrain edge points
    
    Version: 3.5
    Datum 20200508
    Door: Jan den Daas
    Made compatible with sea database in WGS1984
   
"""
import bim6_versioninfo
scriptversion = bim6_versioninfo.version
scriptyear = bim6_versioninfo.year
from generalFunctions import *
import sys, os, string, math
import arcpy
arcpy.env.overwriteOutput = 1

######################
### definities
######################
nrDecimals = 9
inBetweenCheck = 0.01

class PointData:
    def __init__(self, x, y, num):
        self.X = x
        self.Y = y
        self.Nummer = num
        
    def __hash__(self):
        return hash((self.X, self.Y, self.Nummer))
 
    def __eq__(self, other):
        return (self.X, self.Y,self.Nummer) == (other.X, other.Y,self.Nummer)            

        
class ElevationData(PointData):
    def __init__(self, x, y, num, k, kb,lh=-9999, rh=-9999, 
                       th=-9999, tl=4, tr=4, ruw=0):
        PointData.__init__(self, x, y, num)
        self.Hoogte = k
        self.Linkerhoogte = lh
        self.Rechterhoogte = rh
        self.Teenhoogte = th
        self.Ruwheid = ruw
        self.Kruinbreedte = kb
        self.Taludlinks = tl
        self.Taludrechts = tr

        
class BridgeData(PointData):
    def __init__(self, x, y, num, diam=5, cp=1, char='', name=''):
        PointData.__init__(self, x, y, num)
        self.Diameter = diam
        self.CP = cp
        self.Characteristics = char
        self.Name = name
        

def featCount(features):
    result = arcpy.GetCount_management(features)
    count = int(result.getOutput(0))
    del result
    return count

def defineAngle(x0, x1, y0, y1):
    distX = x1 - x0
    distY = y1 - y0
    A = math.atan2(distY,distX)
    return A

def defineLength(x0,x1,y0,y1):
    distX = x1 - x0
    distY = y1 - y0
    return math.sqrt((distX**2 + distY**2))

def loodrechtX(ang, dist):
    return dist * math.sin(ang)

def loodrechtY(ang, dist):
    return dist * math.cos(ang)

 
fieldDict = {"terrain_jump_points": ["Shape@", "NUMBER",  "ROUGHNESS_CODE",
                                     "ELEVATION_LEFT", "ELEVATION_RIGHT",
                                     "SLOPE_LEFT", "SLOPE_RIGHT","CREST_WIDTH","ELEVATION_CREST"],
                    "elevated_line_points": ["Shape@", "NUMBER",  "ROUGHNESS_CODE",
                                     "ELEVATION_LEFT", "ELEVATION_RIGHT",
                                     "SLOPE_LEFT", "SLOPE_RIGHT","CREST_WIDTH","ELEVATION_CREST"],
               "terrain_edge_points": ["Shape@", "NUMBER", "ELEVATION_CREST"],
               "flow_blocking_points": ["Shape@", "NUMBER", "ELEVATION_CREST"],
               "flow_blocking_polygons_points": ["Shape@", "NUMBER", "ELEVATION_CREST"],
               "bridge_points": ['Shape@', 'NUMBER', 'DIAMETER', 'CP', 'CHARACTERISTICS', 'NAME']
            }

def placeElevationPoint(x, y, num, kruinhoogte, fieldDictName, crestWidth, currentPoint=None):  
    if currentPoint:
        teenhoogteDict = {"ROUGHNESS_CODE": currentPoint.Ruwheid,
                          "ELEVATION_LEFT" : currentPoint.Linkerhoogte,
                          "ELEVATION_RIGHT" : currentPoint.Rechterhoogte,
                          "TEENHOOGTE": currentPoint.Teenhoogte,
                          "SLOPE_LEFT" : currentPoint.Taludlinks,
                          "SLOPE_RIGHT" : currentPoint.Taludrechts,
                          "CREST_WIDTH" : currentPoint.Kruinbreedte
                          }
    else:
        teenhoogteDict = {"ROUGHNESS_CODE": 0,
                          "ELEVATION_LEFT" : -9999,
                          "ELEVATION_RIGHT" : -9999,
                          "TEENHOOGTE": -9999,
                          "SLOPE_LEFT" : 4,
                          "SLOPE_RIGHT" : 4,
                          "CREST_WIDTH" : crestWidth,
                          }

    fields = fieldDict[fieldDictName.lower()]
    data = [None]*len(fields)
    newPoint = arcpy.Point(X=x, Y=y)

    for i in range(0,len(fields)):
        name = fields[i]
        if name == "NUMBER":
            data[i] = num
        elif name == "ELEVATION_CREST":
            data[i] = kruinhoogte
        elif name == "Shape@":
            data[i] = newPoint
        elif name in teenhoogteDict:
            data[i] = teenhoogteDict[name] 
    return data

def placeBridgePoint(x, y, num, diam, cp, fieldDictName, char, name, currentPoint=None):  
    if currentPoint:
        teenhoogteDict = {"DIAMETER" : currentPoint.Diameter,
                          "CP" : currentPoint.CP,
                          'CHARACTERISTICS': currentPoint.Characteristics,
                          'NAME': currentPoint.Name}
    else:
        teenhoogteDict = {"DIAMETER" : diam,
                          "CP" : cp,
                          "CHARACTERISTICS": char,
                          'NAME': name}

    fields = fieldDict[fieldDictName.lower()]
    data = [None]*len(fields)
    newPoint = arcpy.Point(X=x, Y=y)

    for i in range(0,len(fields)):
        name = fields[i]
        if name == "NUMBER":
            data[i] = num
        elif name == "Shape@":
            data[i] = newPoint
        elif name in teenhoogteDict:
            data[i] = teenhoogteDict[name] 
    
    return data    
    

def createFinalFc(workspace, pointFC):
    scriptGDB = GetBaselineScriptTemplate()
    arcpy.CopyFeatures_management(os.path.join(scriptGDB, pointFC), os.path.join(workspace, pointFC))

def booleanizeInput(inputParameter):
    if inputParameter == True or inputParameter == False:
        return inputParameter
    elif str(inputParameter) == '1' or str(inputParameter.lower()) == 'true':
        return True
    else:
        return False   

def AddMessage(message):
    print message
    arcpy.AddMessage(message)

def AddError(message):
    print 'Error: ' + message
    arcpy.AddError(message)

def AddWarning(message):
    print 'Warning: ' + message
    arcpy.AddWarning(message)

def checkFieldExists(fc,fieldName):
    flds = arcpy.ListFields(fc)
    for fld in flds:
        if fld.name.lower() == fieldName.lower():
            del flds
            return True
    return False

def deleteExistingPoints(puntFc,nummer,overwritePoints):
    lyr = "countLayer"
    arcpy.MakeFeatureLayer_management(puntFc, lyr, ("\"NUMBER\" = %s" % nummer))
    count = featCount(lyr)
    if count > 0: # Dan bestaat deze al
        #Verwijderen van bestaande punten
        if overwritePoints == True:
            arcpy.DeleteFeatures_management(lyr)
            arcpy.Delete_management(lyr)
            return True
        else:
            arcpy.Delete_management(lyr)
            return False
            
def is_between(a,b, c):
    p = (b.X - a.X)
    q = (c.Y - a.Y)
    r = (c.X - a.X)
    s = (b.Y - a.Y)
    t = cmp(a.X, c.X)
    u = cmp(b.X, c.X)
    v = cmp(a.Y, c.Y)
    w = cmp(b.Y, c.Y)
    m = p * q
    n = r * s
    l = abs(m - n)
    return (l < inBetweenCheck and 
            abs(t + u) <= 1 and
            abs(v + w) <= 1)            

def placePointOnLine(inLineFeatures, overwritePoints, hoogteofdiameter, interval=None, template="", cp=0, isBridge=False, crestWidth=None):
    if interval == 0:
        interval = None
    inLineDesc = arcpy.Describe(inLineFeatures)
    inLinePath = inLineDesc.CatalogPath
    inLineFcName = os.path.basename(inLinePath)
    inLineShapeFieldName = inLineDesc.ShapeFieldName

    AddMessage("Script: %s" % os.path.basename(sys.argv[0]))
    AddMessage("Version: %s" % scriptversion)
    AddMessage("WSP / Deltares, %s\n" % scriptyear)
    AddMessage("Line location: %s" % inLinePath)

    sea = False
    if arcpy.Describe(inLineFeatures).spatialReference.name == "GCS_WGS_1984":
        sea=True
    



    if inLinePath.endswith("_routes"):
        AddError("""Routes en events must be converted to lines and points prior to further processing.
                     Use the tool Routes & Events to Lines & Points.""")
        return #Stoppen
    elif inLinePath.endswith("_lines"):
        outPointPath = os.path.basename(inLinePath.replace("_lines","_points"))

    inLineCur = arcpy.da.SearchCursor(inLineFeatures, "NUMBER")
    numList = []
    for row in inLineCur:
        nummer = row[0]
        if nummer is None:
            AddError("For one or more lines NUMBER is None/NULL. It should be an unique positive integer. Fix this before running this tool again. Process aborted.")
            return            
        if nummer in numList:
            AddError("One or more lines does not have an unique NUMBER. Each number should an unique positive integer. Fix this before running this tool again. Process aborted.")
            return
        else:
            numList.append(nummer)
    del inLineCur, numList
                 
    outPointPath = os.path.join(os.path.dirname(inLinePath),outPointPath)
    outPointWs = os.path.dirname(outPointPath)
    outPointName = os.path.basename(outPointPath)

    if template != "":
        templateName = template + "_points"
    else:
        templateName = outPointName



    #Check of puntenbestand bestaat
    #Zo ja: check welke velden er zijn
    AddMessage("Point location: %s" % outPointPath)
    
    keepExistingPoints = False
    if arcpy.Exists(outPointPath):
        #Als punten overschrijven aanstaat
        #Eerst maar even kijken of alle lijnen geselecteerd zijn
        #Zo ja: donder dan het puntenbestand weg en beschouw het als een volledig
        # nieuwe opbouw
        if overwritePoints == True:
            arcpy.MakeFeatureLayer_management(inLinePath, "allFeat")
            if featCount("allFeat") == featCount(inLineFeatures):
                arcpy.Delete_management(outPointPath)
                createFinalFc(outPointWs, outPointName)
            else:
                AddMessage("Overwriting existing Points")
                deletelist=[]
                with arcpy.da.SearchCursor(inLineFeatures, "NUMBER") as inlinecur:
                    for row in inlinecur:
                        deletelist.append(row[0])
                with arcpy.da.UpdateCursor(outPointPath, "NUMBER") as pointcur:
                    for row in pointcur:
                        if row[0] in deletelist:
                            pointcur.deleteRow()
            arcpy.Delete_management("allFeat")
        else:
            keepExistingPoints = True
    else:
        createFinalFc(outPointWs, outPointName)
        
    # Loop door de lijn 
    # If else: afh van of bestaande punten behouden moeten blijven
    if keepExistingPoints == False:
        AddMessage("Determine location of points...")
        if isBridge:
            inLineCur = arcpy.da.SearchCursor(inLineFeatures, ["Shape@", "NUMBER", 'CHARACTERISTICS','NAME'])

        else:
            inLineCur = arcpy.da.SearchCursor(inLineFeatures, ["Shape@", "NUMBER"])
        
        #pointCur = punten cursor voor wegschrijven
        pointCur = arcpy.da.InsertCursor(outPointPath, fieldDict[templateName.lower()])

        #Plaats tussenliggende punten, voor een zee database
        if interval != None and sea:
            sealines = arcpy.SplitLine_management(inLineFeatures, "in_memory/sealines")
            seapoints = arcpy.GeneratePointsAlongLines_management(sealines, "in_memory/seapoints", "DISTANCE", "{} Meters".format(interval))
            if isBridge:
                seaCur = arcpy.da.SearchCursor(seapoints, ["Shape@X","Shape@Y", "NUMBER", 'CHARACTERISTICS','NAME'])
            else:
                seaCur = arcpy.da.SearchCursor(seapoints, ["Shape@X","Shape@Y", "NUMBER"])
                
            for seapoint in seaCur:
                if isBridge:
                    row = placeBridgePoint(seapoint[0], seapoint[1], seapoint[2], hoogteofdiameter, cp, templateName, inLineFeat[2], inLineFeat[3])
                else:
                    row = placeElevationPoint(seapoint[0], seapoint[1], seapoint[2], hoogteofdiameter,templateName, crestWidth)
                pointCur.insertRow(row)

        lineList = {}
        for inLineFeat in inLineCur:
            feat = inLineFeat[0]
            featLength = feat.length
            nummer = inLineFeat[1]
    
            #Controleer of lijn multipart is. Dit mag niet
            if feat.partCount > 1:
                AddMessage("Input line with NUMBER = %s contains multiparts. No points are made for this line." %nummer)
                continue
                
            #Startwaarden voor bepaling puntlocaties
            boolFirstPoint = True
            boolSecondPoint = True
            vertex = 1
            vertexCount = feat.pointCount
            x0 = -999999
            y0 = -999999
            
            #Loop door alle vertices in de lijn
            prt = feat.getPart(0)
       
            
            lineArray = []
            vertPnt = arcpy.CreateObject("Point")
            for pnt in prt:
                if boolFirstPoint == True:
                    boolFirstPoint = False
                else:
                    angle = defineAngle(x0, pnt.X, y0, pnt.Y)
                    length = defineLength(x0, pnt.X, y0, pnt.Y)
                    if boolSecondPoint == True:
                        boolSecondPoint = False
                        vertPnt.X = x0
                        vertPnt.Y = y0
                        lineArray.append([x0, y0])
                        if isBridge:
                            row = placeBridgePoint(x0, y0, nummer, hoogteofdiameter, cp, templateName, inLineFeat[2], inLineFeat[3])
                        else:
                            row = placeElevationPoint(x0, y0, nummer, hoogteofdiameter,templateName, crestWidth)
                        pointCur.insertRow(row)
                        vertex+=1
    
                    #Plaats tussenliggende punten, maar alleen als er nieuwe punten gemaakt moeten worden
                    if interval != None and sea == False:
                        cumDist = interval
                        while cumDist < round(length,3):
                            perc = cumDist / length
                            subX = x0 + perc * (pnt.X - x0)
                            subY = y0 + perc * (pnt.Y - y0)
                            vertPnt.X = subX
                            vertPnt.Y = subY
                            lineArray.append([subX, subY])
                            if isBridge:
                                row = placeBridgePoint(subX, subY, nummer, hoogteofdiameter, cp, templateName, inLineFeat[2], inLineFeat[3])
                            else:
                                row = placeElevationPoint(subX,subY,nummer, hoogteofdiameter,templateName, crestWidth)
                            pointCur.insertRow(row)
                            cumDist += interval
                            vertex+=1                
                    
                    #Plaats punt op eindvertex lijndeel
                    vertPnt.X = pnt.X
                    vertPnt.Y = pnt.Y
                    lineArray.append([pnt.X, pnt.Y])
                    if isBridge:
                        row = placeBridgePoint(pnt.X,pnt.Y, nummer, hoogteofdiameter, cp, templateName, inLineFeat[2], inLineFeat[3])
                    else:                    
                        row = placeElevationPoint(pnt.X,pnt.Y,nummer, hoogteofdiameter,templateName,crestWidth)
                    pointCur.insertRow(row)
                    vertex+=1
                x0 = pnt.X
                y0 = pnt.Y

            lineList[nummer] = lineArray
            del prt         
        del pointCur

        inLineCur = arcpy.da.UpdateCursor(inLineFeatures, ["Shape@", "NUMBER"])

        for row in inLineCur:
            arr = lineList[row[1]]
            row[0] =  arr          
            inLineCur.updateRow(row)

        del inLineCur
        
    else:
        AddMessage("Keeping existing points")
        if isBridge:
            inLineCur = arcpy.da.SearchCursor(inLineFeatures, ["Shape@", "NUMBER", 'CHARACTERISTICS', 'NAME'])
        else:
            inLineCur = arcpy.da.SearchCursor(inLineFeatures, ["Shape@", "NUMBER"])

        dictLines = {}    
        dictKenmerk = {}
        dictBridgeNames = {}
        
        for inLineFeat in inLineCur:
            
            feat = inLineFeat[0]
            featLength = feat.length
            nummer = inLineFeat[1]        
            if isBridge:
                dictKenmerk[nummer] = inLineFeat[2] 
                dictBridgeNames[nummer] = inLineFeat[3]
    
            #Controleer of lijn multipart is. Dit mag niet
            if feat.partCount > 1:
                AddMessage("Input line with NUMBER = %s contains multiparts. No points are made for this line." %nummer)
                inLineFeat = inLineCur.next()
                continue
            
            dictLines[nummer] = []
            
            prt = feat.getPart(0)
            for pnt in prt:
                if isBridge:
                    pd = BridgeData(round(pnt.X, nrDecimals), round(pnt.Y, nrDecimals), nummer, hoogteofdiameter, cp, dictKenmerk[nummer], dictBridgeNames[nummer])
                else:
                    pd = ElevationData(round(pnt.X,nrDecimals), round(pnt.Y,nrDecimals),nummer, hoogteofdiameter, kb=crestWidth)
                dictLines[nummer].append(pd)
                
        isWeir = False
        if isBridge:
            pointCur = arcpy.da.SearchCursor(outPointPath, fieldDict.get('bridge_points'))   
        elif 'terrain_jump_points' in os.path.basename(outPointPath).lower():
            pointCur = arcpy.da.SearchCursor(outPointPath, fieldDict.get('terrain_jump_points'))
            isWeir = True 
        elif 'elevated_line_points' in os.path.basename(outPointPath).lower():
            pointCur = arcpy.da.SearchCursor(outPointPath, fieldDict.get('elevated_line_points'))
            isWeir = True                                                                    
        else:
            pointCur = arcpy.da.SearchCursor(outPointPath, fieldDict.get('terrain_edge_points'))
        useSelection = False
        
        for pointRow in pointCur:
            feat = pointRow[0]
            nummer = pointRow[1]
            if not nummer in dictLines:
                useSelection = True
                continue
            
            pnt = feat.getPart()
            
            if isBridge:
                diam = pointRow[2]
                cp = pointRow[3]
                pd = BridgeData(round(pnt.X, nrDecimals), round(pnt.Y, nrDecimals), nummer, diam, cp, dictKenmerk[nummer], dictBridgeNames[nummer])
            else:
                if 'terrain_edge_points' in os.path.basename(outPointPath).lower():
                    kh = pointRow[2]
                else:
                    kh = pointRow[8]
                if isWeir:
                    pd = ElevationData(round(pnt.X, nrDecimals), round(pnt.Y, nrDecimals), nummer, kh, lh=pointRow[3],
                                    rh=pointRow[4], tl=pointRow[5], tr=pointRow[6], kb=pointRow[7], ruw=pointRow[2])
                else:
                    pd = ElevationData(round(pnt.X, nrDecimals), round(pnt.Y, nrDecimals), nummer, kh, kb=crestWidth) 
            
            
            lineCoords = dictLines[nummer]
            
            # de check of pd in lineCoords zit is op basis van X, Y en nummer
            if not pd in lineCoords: 
                for index in range(1, len(lineCoords)):
                    firstPoint = lineCoords[index - 1]
                    lastPoint = lineCoords[index]
                    if is_between(firstPoint, lastPoint, pd):
                        lineCoords.insert(index, pd)
                        break
            else:
                allIndices = [i for i, x in enumerate(lineCoords) if x == pd]
                for index in allIndices:                
                    lineCoords[index] = pd

        del pointCur
        
        AddMessage("Recreating lines...")       
        inLineCur = arcpy.da.UpdateCursor(inLineFeatures, ["Shape@", "NUMBER"])

        for inLineFeat in inLineCur:
            nummer = inLineFeat[1]
            lineArray = []   
            
            coords = dictLines[nummer]
            
            for coord in coords:
                lineArray.append([coord.X, coord.Y])

            inLineFeat[0] = lineArray
            inLineCur.updateRow(inLineFeat)        
        del inLineCur
        
        if useSelection == False:
            arcpy.DeleteFeatures_management(outPointPath)
        else:
            strList = [str(num) for num in dictLines]            
            strSelection = (""" "NUMBER" IN (%s)""" % (", ".join(strList)))
            lyr = "lyr" 
            arcpy.MakeFeatureLayer_management(outPointPath, lyr, strSelection)
            arcpy.DeleteFeatures_management(lyr)
            arcpy.Delete_management(lyr)
            
        inPointCur = arcpy.da.InsertCursor(outPointPath, fieldDict[templateName])
  
        for nummer in dictLines:
            coords = dictLines[nummer]
            
            for coord in coords:
                pnt.X = coord.X
                pnt.Y = coord.Y
                if isBridge:
                    pointRow = placeBridgePoint(coord.X, coord.Y, nummer, coord.Diameter, coord.CP, templateName, dictKenmerk[nummer], dictBridgeNames[nummer], currentPoint=coord)
                else:
                    pointRow = placeElevationPoint(coord.X,coord.Y,nummer, coord.Hoogte,templateName, crestWidth=crestWidth, currentPoint=coord)
                inPointCur.insertRow(pointRow)
        del inPointCur        


######################
### hoofdscript
######################



if __name__ == "__main__":
    inLineFeatures = arcpy.GetParameterAsText(0)
    overwritePoints = arcpy.GetParameter(1)
    if "bridge" in inLineFeatures:         
        # Brugpunten prikker
        diameter = arcpy.GetParameter(2)
        cp = arcpy.GetParameter(3)
        interval = None
        placePointOnLine(inLineFeatures,overwritePoints,diameter,interval, cp=cp, isBridge=True)          
    elif "elevated" in inLineFeatures:
        hoogte = arcpy.GetParameter(2)
        interval = arcpy.GetParameter(3)
        if interval == 0 :
            interval = None
        placePointOnLine(inLineFeatures,overwritePoints,hoogte,interval, isBridge=False, crestWidth=3)
    else:
        hoogte = arcpy.GetParameter(2)
        interval = arcpy.GetParameter(3)
        if interval == 0 :
            interval = None
        placePointOnLine(inLineFeatures,overwritePoints,hoogte,interval, isBridge=False, crestWidth=10)      


    