# -*- coding: cp1252 -*-
'''
    Script: bim5_KruinhoogtePrikker.py
    Versie: zie script bim5_versioninfo.py

    Doel: prikken kruinhoogtes uit TIN

    Auteur: Koen Volleberg, Lievense
    Voor: Stichting Deltares

    Updates:
    Versie: 1
    Datum: 13112014
    Door: Koen Volleberg
    Wat: eerste volledige release

    Versie: 2.0
    Datum: 19052015
    Door: Koen Volleberg
    Wat: optimalisatie naar ArcGIS 10.1
    - betere error meldingen
    
    Version: 2.1
    Datum 20171127
    Door: Niels Mulder
    Update to Baseline 6
    
    Version: 2.2
    Datum 20181205
    Door: Jan den Daas
    Default NoData is -9999
    
    Version: 2.3
    Datum 20221101
    Door: Jan den Daas
    Fix for if elevation is exactly 0.0

'''
import bim6_versioninfo
scriptversion = bim6_versioninfo.version
scriptyear = bim6_versioninfo.year

import os, sys
import arcpy
import datetime
import traceback
arcpy.env.overwriteOutput = 1

#Global settings
#veldnamen
kField = "KRUINHOOGTE"
nmField ="NUMMER"
kmField=  "KENMERK"
useStandardDistance = False

##########
## classes
##########

class PointData:
    def __init__(self, pnt, num, vert,k,km):
        self.Point = pnt
        self.Nummer = num
        self.Vertex = vert
        self.Kruin = k
        self.Kenmerk = km

class SpecificException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)
    

##########
## defs
##########

"""
def featCount(f):
    result = arcpy.GetCount_management(f)
    count = int(result.getOutput(0))
    return count
"""

def AddMessage(message):
    print message
    arcpy.AddMessage(message)

def AddError(message):
    print 'Error: ' + message
    arcpy.AddError(message)

def AddWarning(message):
    print 'Warning: ' + message
    arcpy.AddWarning(message)

def getOidFieldName(fc):
    desc = arcpy.Describe(fc)
    oid = desc.OidFieldName
    del desc
    return oid

"""
def getShapeFieldName(fc):
    desc = arcpy.Describe(fc)
    nm = desc.ShapeFieldName
    del desc
    return nm
"""

def PrikKruinhoogtes(lijnen, surf, changeAllElevation, elev):
    AddMessage("Script: %s" % os.path.basename(sys.argv[0]))
    AddMessage("Version: %s" % scriptversion)
    AddMessage("Lievense / Deltares, %s\n" % scriptyear)
    
    try:
        validFilenames = ["terrain_jump_", "elevated_line_", "terrain_edge_"]
        isValid = False
        for fn in validFilenames:
            if fn in lijnen.lower():
                isValid = True

        if not arcpy.Exists(surf):
            raise SpecificException("No valid elevation model selected. Process aborted")  

        if isValid == False:
            raise SpecificException("No valid Baseline feature class selected. Process aborted")

        ds = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        
        if arcpy.CheckExtension("3D") == "Available":
            arcpy.CheckOutExtension("3D")
        else:
            raise SpecificException("3D Analyst is not available. Process aborted")

        baseDir = os.path.dirname(os.path.dirname(os.path.dirname(lijnen)))

        surfaceDesc = arcpy.Describe(surf)
        dsType = surfaceDesc.datasetType.lower()
        tempTin = None
        if dsType not in ['terrain', 'tin', 'rasterdataset']:
            raise SpecificException("No valid elevation model selected.")
##        elif dsType == 'terrain':
##            tempTin = os.path.join(baseDir,"temptin")
##            arcpy.TerrainToTin_3d(surf, tempTin, pyramid_level_resolution=0,
##                               max_nodes=500000000, clip_to_extent="CLIP")

        #Information and configure output names
        lijnenName = os.path.basename(lijnen)
        lijnenName = lijnenName.replace("_lines","")
        inLineDesc = arcpy.Describe(lijnen)
        inLinePath = inLineDesc.catalogPath
        inLineShapeFieldName = inLineDesc.ShapeFieldName
        punten = inLinePath.replace("_lines","_points")

        lineCur = arcpy.da.SearchCursor(lijnen, ["NUMBER"])

        updateNummers = []
        for line in lineCur:
            num = line[0]
            updateNummers.append(num)
        del lineCur
            
        arcpy.AddField_management(punten,"UNIQUEID","LONG")
        #arcpy.CalculateField_management(punten,"UNIQUEID",("!%s!" %getOidFieldName(punten)),"PYTHON_9.3")
        arcpy.CalculateField_management(punten,"UNIQUEID",("!%s!" %arcpy.Describe(punten).OidFieldName),"PYTHON_9.3")

        inmemPoints = 'in_memory/points'
        arcpy.CopyFeatures_management(punten,inmemPoints)
        
        AddMessage("Get elevation...")
##        if tempTin != None:
##            arcpy.TerrainToTin_3d(surf,tempTin,"0","500000000","True")
##            arcpy.AddSurfaceInformation_3d(tempTin, inmemPoints, "Z")
##            arcpy.Delete_management(tempTin)
##        else:
        arcpy.AddSurfaceInformation_3d(in_surface=surf,
                                       in_feature_class=inmemPoints,
                                       out_property="Z")

        AddMessage("Read through elevation values...")
        dictUpdate = {}
        pointCur = arcpy.da.SearchCursor(inmemPoints, ["Z", "UNIQUEID"])

        for pointRow in pointCur:
            z = pointRow[0]
            un = pointRow[1]
            dictUpdate[un] = z
        
        del pointCur

        AddMessage("Update elevation for selected lines...")
        uCur = arcpy.da.UpdateCursor(punten, ["NUMBER", "ELEVATION_CREST", "UNIQUEID", 'OID@'])
        
        for uRow in uCur:
            num = uRow[0]
            if not num in updateNummers:
                continue
            else:
                if not changeAllElevation:
                    if uRow[1] != elev:
                        continue
                un = uRow[2]
                z = dictUpdate[un]
                
                if not z and z!=0:
                    z = -9999
                    AddWarning('Z value for point with ObjectID = {0} could not be determined. Z-value set to -9999 Check manually.'.format(uRow[3]))
                uRow[1] = z
                uCur.updateRow(uRow)
        del uCur
        
        arcpy.DeleteField_management(punten,"UNIQUEID")
        arcpy.Delete_management(inmemPoints)
    except SpecificException, error:
        AddError(error.value)
    except:
        tb = sys.exc_info()[2]
        tbinfo = traceback.format_tb(tb)[0]

        err = "Traceback info:\n" + tbinfo + "\nError Info:\n" + str(sys.exc_info()[1])      
        AddError(err)
    finally:
        arcpy.CheckInExtension("3D")
##############################################################################################
##
##############################################################################################    

if __name__ == "__main__":
    lijnen = arcpy.GetParameterAsText(0)
    surf = arcpy.GetParameterAsText(1)
    changeAllElevation = arcpy.GetParameter(2)
    elev = arcpy.GetParameter(3)
    

    #lijnen = 'C:/baseline/spelen/testen/testset_b5/myBL6collection/measures_baseline6/maatregelen/nr_doorwcl_a2/baseline.gdb/elevation/terrainjumps_lines'
    #surf = 'C:/baseline/spelen/testen/testset_b5/myBL6collection/ref/baseline.gdb/elevation/elevation_model_terrain'
    #changeAllElevation = True
    #elev = None

    
    PrikKruinhoogtes(lijnen, surf, changeAllElevation, elev)
